
//		EKeys.cpp - Hotkeys to edit words on the command line


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define PLUGIN_NAME					L"EKeys"

#define VER_MAJOR					0
#define VER_MINOR					51
#define VER_BUILD					0
#define VER_PATCH					3

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif


#define MAX_ARG_LEN					8192
#define MAX_ARGS					1024
#define SHORT_BUF_LEN				1024
#define MAX_CMD_LINE_LEN			16384
#define MAX_FILENAME_LEN			32768


#define FLAG_EXPAND_VARIABLES		0x0001
#define FLAG_AUTO_ENTER				0x0002
#define FLAG_STRIP_QUOTES			0x0004
#define FLAG_AUTOQUOTE				0x0008
#define FLAG_COMMAND_PREFIX			0x0010
#define FLAG_NO_ERROR_BEEP			0x0020
#define FLAG_INSERT					0x0040
#define FLAG_CHANGE_OPTION			0x1000
#define FLAG_ALL_DEFINED			0x107f

#define PARAM_CHAR					'#'


#define ENTER						0x0d
#define SPACE						0x20
#define QUOTE						0x22
#define NBSP						0xa0
#define ZWNJ						0x200c
#define BOM							0xfeff
#define SAFECHAR_DIFF				0xfee0
#define FULLWIDTH_EQUALS			0xff1d

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )



HANDLE MyDLLHandle = NULL;


long int ShellMajor = -1, ShellMinor = -1, ShellBuild = -1;



#include "hotkeys.cpp"



//	Structure used to record a user-defined keystroke definition:

struct _KeyEntry {
	_KeyEntry *Next;				//	pointer to the next entry in the list
	int Key;						//	key value
	unsigned int Flags;				//	bitmapped flags
	wchar_t Text;					//	replacement text
};



_KeyEntry *FirstKey		= NULL;			//	Pointer to the first keystroke definition
BOOL Debug				= FALSE;		//	Plugin in debug mode


const wchar_t PluginFeatures[] = L"ekey,*key";


const wchar_t HelpWindowTitle[] = L"EKeys Plugin Help";


const wchar_t EKeyHelpText[] = L"Define, list, or remove hotkeys.\n\
\n\
EKEY /B /C /D /E /I /L /O /Q /R:file /S:file /T /U /X hotkey=text  \r\n\
\n\
\t/B   no error Beep\n\
\t/C   Command at start of text\n\
\t/D   Delete a hotkey definition\n\
\t/E   automatically Execute\n\
\t/I   Insert text if cursor not at a word\n\
\t/L   List all supported hotkeys\n\
\t/O   Overwrite existing file\n\
\t/Q   Quote resulting text if needed\n\
\t/R:  Read key definitions from a file\n\
\t/S:  Save key definitions to a file\n\
\t/T   change TCC option\n\
\t/U   Unquote word\n\
\t/X   expand variables and functions\n\
\n\
EKey allows you to define custom hotkeys, which operate on the space-  \n\
delimited 'word' at the cursor.  When you press a hotkey, the word at\n\
the cursor is replaced with the text in the key\u2019s definition; a number  \n\
sign # in the text represents the original word. \n";


/*
const wchar_t F_EKeyHelpText[] = L"@EKEY[hotkey]  \r\n\n\
Returns the definition text for the given hotkey.  \n";

const wchar_t _EKeysHelpText[] = L"_EKEYS  \r\n\n\
Returns the number of hotkeys defined.  \n";
*/


const wchar_t ClipFilename[]		= L"Clip:";
const wchar_t StdinFilename[]		= L"-";
const wchar_t AutoLoadFilename[]	= L"Auto.EK";




//	Display the start of an error message to stderr :

int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}


#include "ParseArgs.cpp"


//	Deletes all items in a linked list, and zeros the pointer to the first item :

int FreeLinkedList( void **Pointer )
{
	if ( Pointer == NULL )
		return -1;

	size_t *This = (size_t*) *Pointer, Next = NULL;

	while ( This ) {
		// Printf( L"   * Freeing memory at 0x%08x\r\n", This );
		Next = *This;
		free( This );
		This = (size_t*) Next;
	}

	*Pointer = NULL;
	return 0;
}



//	Expands and parses %_4VER :

int GetShellVersion()
{
	wchar_t Temp[SHORT_BUF_LEN]		= L"%_4ver";
	wchar_t *sep1					= NULL;
	
	ExpandVariables( Temp, 1 );
	ShellMajor = wcstol( Temp, &sep1, 10 );

	if ( *sep1 ) {
		sep1++;
		if ( *sep1 )
			ShellMinor = wcstol( sep1, &sep1, 10 );
	}

	wcscpy_s( Temp, SHORT_BUF_LEN, L"%_build" );
	ExpandVariables( Temp, 1 );
	ShellBuild = wcstol( Temp, NULL, 10 );

	if ( ( ShellMajor < 8 ) || ( ShellMajor > 99 ) || ( ShellMinor < 0 ) || ( ShellMinor > 99 ) || ( ShellBuild < 0 ) || ( ShellBuild > 9999 ) )
		ShellMajor = ShellMinor = ShellBuild = -1;

	return ShellMajor;
}



BOOL IsAFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD attr		= GetFileAttributes( Filename );


	if ( attr == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( attr & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



//	DLL entry point :

BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
			MyDLLHandle = hModule;
			break;

		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/ekeys.html";
	piInfo.pszDescription	= (LPTSTR) L"Hotkeys to edit words on the command line";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



int AutoLoadKeys();



//	Initializes the plugin when it is loaded :

DLLExports BOOL WINAPI InitializePlugin( void )
{
	InitHotkeys();


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	GetShellVersion();
	AutoLoadKeys();
	return 0;
}



//	Called when the plugin is unloaded.
//	Frees the linked list of key definitions.

DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	FreeLinkedList( (void**) &FirstKey );
	return 0;
}



//	Attempts to convert a Windows error code to a descriptive string

int GetErrorText( int err, LPTSTR outString )
{
	if ( outString == NULL )
		return -1;

	if ( err == -1 )		//	I use -1 a lot to indicate internal errors like null pointers, etc.
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



//	Does inHandle refer to the console?

int ConsoleHandle( HANDLE inHandle )
{
	if ( inHandle == NULL )
		return 0;

	if ( ( inHandle == (HANDLE) STD_INPUT_HANDLE ) || ( inHandle == (HANDLE) STD_OUTPUT_HANDLE ) || ( inHandle == (HANDLE) STD_ERROR_HANDLE ) )
		inHandle = GetStdHandle( (DWORD) inHandle );

	DWORD Mode = UINT_MAX;
	int rv = GetConsoleMode( inHandle, &Mode );

	if ( ( rv == 0 ) || ( Mode == UINT_MAX ) )
		return 0;
	else
		return 1;
}



int ConsoleHandle( DWORD inHandle )
{
	return ConsoleHandle( (HANDLE) inHandle );
}



//	Some memory-mapped file handling code which I use to load the key definitions from
//	a file.  It supports UTF-16, UTF-8, OEM code pages, the clipboard, and stdin from
//	the console and pipes.

#include "codepages.cpp"
#include "mmfiles.cpp"



//	Nasty kludge to check for "alt-" or "a-" at the start of a string.  This is to deal with
//	the keynames "alt-=" and "a-=", in which parsing should break at the SECOND equals sign,
//	not the first.  Yes, this is inelegant.

int _NotAlt( LPTSTR String )
{
	if ( !String )
		return -1;

	if ( !_wcsicmp( String, L"alt-" ) )
		return 0;

	if ( !_wcsicmp( String, L"a-" ) )
		return 0;

	return 1;
}



//	Often in options, I treat an equals sign the same as a colon.  This can be handy
//	when doing tab completion on options which take a filename; TCC's tab completion
//	mechanism breaks arguments at an unquoted equals sign.

BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



//	Parses an unsigned integer from inString, and stores it in *outValue.  The
//	number may be either decimal, or hex with a leading 0X.  Returns 0 for
//	success; on an error, returns nonzero and does not change *outValue.  If
//	EndChar is not null, saves a pointer to the next character after the number
//	in *EndChar; if EndChar _is_ null, then the value must be null-terminated or
//	this routine will flag an error.

int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int ReplaceChar( LPTSTR String, wchar_t From, wchar_t To )
{
	if ( !String )
		return -1;

	if ( !From || !To || ( From == To ) )
		return 0;


	int k		= 0;


	for ( wchar_t *ch = String; *ch; ch++ )
		if ( *ch == From )
			*ch = To, k++;

	return k;
}



//	Displays help text, and optionally plugin version info, to stdout :

int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



//	Parses /B /C /E /I /Q /U /X (which may be mashed together).

int ParseFlagsOption( LPTSTR inString, unsigned int *outFlags )
{
	if ( !inString || !outFlags )
		return -1;

	if ( inString[0] != '/' )
		return 1;


	size_t i = 1;
	unsigned int Flags = 0;


	while ( inString[i] ) {
		switch ( towupper( inString[i] ) ) {

		case 'B' :
			Flags |= FLAG_NO_ERROR_BEEP;
			break;

		case 'C' :
			Flags |= FLAG_COMMAND_PREFIX;
			break;

		case 'E' :
			Flags |= FLAG_AUTO_ENTER;
			break;

		case 'I' :
			Flags |= FLAG_INSERT;
			break;

		case 'Q' :
			Flags |= FLAG_AUTOQUOTE;
			break;

		case 'T' :
			Flags |= FLAG_CHANGE_OPTION;
			break;

		case 'U' :
			Flags |= FLAG_STRIP_QUOTES;
			break;

		case 'X' :
			Flags |= FLAG_EXPAND_VARIABLES;
			break;

		case '/' :
			if ( !iswalpha( towupper( inString[i+1] ) ) )
				return 1;

			break;

		default:
			return 1;
			break;
		}

		i++;
	}

	if ( Flags ) {
		*outFlags |= Flags;
		return 0;
	}

	return 1;
}



//	Checks for a string between double quotes, and removes them if found.

int Unquote( LPTSTR String )
{
	if ( !String )
		return -1;


	size_t l = wcslen( String );


	if ( l < 2 )
		return 0;

	if ( ( String[0] != QUOTE ) || ( String[l - 1] != QUOTE ) )
		return 0;

	for ( size_t i = 0; i < ( l - 1); i++ )
		String[i] = String[i + 1];

	String[l - 2] = '\0';
	return 1;
}



//	Is ch a character which should be quoted?

BOOL ProblemChar( wchar_t ch )
{
	if ( iswspace( ch ) )
		return TRUE;

	if ( ( ch == '&' ) || ( ch == ',' ) || ( ch == ';' ) || ( ch == '<' ) || ( ch == '>' ) || ( ch == '=' ) || ( ch == '|' ) )
		return TRUE;

	if ( iswcntrl( ch ) )
		return TRUE;

	return FALSE;
}



//	Checks whether a string contains 'problem' characters, and if so, quotes it.

int AutoQuote( LPTSTR String, size_t BufferSize )
{
	if ( !String )
		return -1;


	BOOL quote = FALSE, NeedsQuotes = FALSE;

	
	for ( size_t i = 0; String[i]; i++ ) {
		if ( String[i] == QUOTE )
			quote = !quote;

		if ( !quote && ProblemChar( String[i] ) )
			NeedsQuotes = TRUE;
	}

	if ( !NeedsQuotes )
		return 0;


	wchar_t *TempBuffer = NULL;

	TempBuffer = (wchar_t *) malloc( ( wcslen( String ) + 3 ) * sizeof( wchar_t ) );
	if ( TempBuffer == NULL )
		return -2;

	size_t j = 0;

	TempBuffer[j++] = QUOTE;

	for ( size_t i = 0; String[i]; i++ )
		if ( String[i] != QUOTE )
			TempBuffer[j++] = String[i];

	TempBuffer[j++] = QUOTE;
	TempBuffer[j] = '\0';

	if ( wcslen( TempBuffer ) >= BufferSize )
		return -3;

	wcscpy_s( String, BufferSize, TempBuffer );
	free( TempBuffer );
	return 1;
}



//	Calls an internal TCC command.

int CallInternal( LPTSTR Cmd, LPTSTR Args )
{
	if ( !Cmd )
		return -1;
	else if ( Cmd[0] == '\0' )
		return -1;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wchar_t ExpVar[SHORT_BUF_LEN]	= L"%_expansion";
	BOOL AddStar					= TRUE;


	ExpandVariables( ExpVar, 1 );
	if ( wcschr( ExpVar, '1' ) )
		AddStar = FALSE;


	if ( ( Cmd[0] != '*' ) && AddStar )
		wcscpy_s( CmdBuf, MAX_ARG_LEN, L"*" );

	wcscat_s( CmdBuf, MAX_ARG_LEN, Cmd );

	if ( Args ) {
		if ( Args[0] ) {
			if ( !iswspace( Args[0] ) )
				wcscat_s( CmdBuf, MAX_ARG_LEN, L" " );

			wcscat_s( CmdBuf, MAX_ARG_LEN, Args );
		}
	}

	return Command( CmdBuf, 0 );
}



//	Adds a filename onto a directory name, adding a slash if needed.

int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



//	Returns the filename, ignoring any drive or path specification.

LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



//	Updates Take Command's status bar :

int StatusBar( LPTSTR Text )
{
	return TakeCommandIPC( L"STATUSBAR", Text );
}



//	Stupid routine to slurp an entire file into memory and slam it onto the clipboard.
//	File is assumed to be OEM encoded unless UTF16 is set, in which case it's UTF-16.

int PutFileOnClipboard( LPTSTR Filename, BOOL UTF16 )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	HANDLE FileHandle		= NULL;
	LARGE_INTEGER FileSize	= { 0, 0 };
	HGLOBAL BufferHandle	= NULL;
	BYTE *Buffer			= NULL;

	DWORD BytesRead			= 0;
	int rv					= 0;


	FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );

	if ( FileHandle == INVALID_HANDLE_VALUE )
		rv = 1;
	else {
		GetFileSizeEx( FileHandle, &FileSize );

		if ( ( FileSize.QuadPart == 0 ) || ( FileSize.QuadPart > 0xffff ) )
			rv = 2;
		else {
			BufferHandle = GlobalAlloc( GHND, FileSize.LowPart + 2 );

			if ( BufferHandle == NULL )
				rv = 3;
			else {
				Buffer = (BYTE*) GlobalLock( BufferHandle );
				
				if ( Buffer == NULL )
					rv = 3;
				else {
					ReadFile( FileHandle, Buffer, FileSize.LowPart, &BytesRead, NULL );

					if ( BytesRead != FileSize.QuadPart )
						rv = 4;
					else {
						Buffer[FileSize.LowPart] = 0x00;
						Buffer[FileSize.LowPart + 1] = 0x00;

						if ( !OpenClipboard( GetConsoleWindow() ) )
							rv = 5;
						else {
							EmptyClipboard();
							SetClipboardData( UTF16 ? CF_UNICODETEXT : CF_OEMTEXT, BufferHandle );
							CloseClipboard();
							Buffer = NULL;
							BufferHandle = NULL;
							rv = 0;
						}
					}

					if ( BufferHandle )
						GlobalUnlock( BufferHandle );
				}

				if ( BufferHandle )
					GlobalFree( BufferHandle );
			}
		}

		CloseHandle( FileHandle );
	}
	
	return rv;
}



//	Saves the linked list of key definitions to a file.
//  Filename :  may be CLIP: (save to the clipboard) or - (dump to stdout)
//	Overwrite:  okay to overwrite an existing file
//	Complain:   display error messages to stderr
//	UTF16:      write the file as UTF-16, else it's UTF-8

int SaveKeysToFile( LPTSTR Filename, BOOL Overwrite, BOOL Complain, BOOL UTF16 )
{
	if ( Filename == NULL )
		return -1;


	HANDLE FileHandle		= NULL;
	BOOL ToClipboard		= FALSE;
	BOOL ToStdout			= FALSE;
	BOOL DeleteWhenDone		= FALSE;
	BOOL WriteBOM			= TRUE;
	DWORD BytesWritten		= 0;
	_KeyEntry *ThisKey		= FirstKey;
	int rv					= 0;
	int error				= 0;

	const int TextSize = MAX_ARG_LEN + SHORT_BUF_LEN;


	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";
	wchar_t KeyString[TextSize] 		= L"";
	BYTE UTF8Buffer[2 * TextSize];



	if ( FirstKey == NULL ) {
		if ( Complain ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Nothing to save!\r\n" );
		}
		return 2; 
	}


	if ( !_wcsicmp( Filename, ClipFilename ) ) {
		ToClipboard = TRUE;
		DeleteWhenDone = TRUE;
		Overwrite = TRUE;
		UTF16 = TRUE;
		WriteBOM = FALSE;

		GetTempPath( MAX_FILENAME_LEN, Filename );				//	GetTempDirectory( Filename );
		GetTempFileName( Filename, L"CLP", 0, Filename );		//	UniqueFileName( Filename, L"CLP" );
	}
	else if ( !wcscmp( Filename, StdinFilename ) ) {
		ToStdout = TRUE;
		UTF16 = TRUE;
		WriteBOM = FALSE;
	}

	memset( UTF8Buffer, 0, sizeof( UTF8Buffer ) );


	if ( !ToStdout )
		FileHandle = CreateFile( Filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, Overwrite ? CREATE_ALWAYS : CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL );


	if ( FileHandle == INVALID_HANDLE_VALUE ) {
		if ( Complain ) {
			GetErrorText( GetLastError(), ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error creating file %s :\r\n", Filename );
			Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		}

		return 2;
	}


	if ( WriteBOM ) {
		if ( UTF16 ) {
			UTF8Buffer[0] = 0xff;
			UTF8Buffer[1] = 0xfe;
			BytesWritten = 2;
		}
		else {
			UTF8Buffer[0] = 0xef;
			UTF8Buffer[1] = 0xbb;
			UTF8Buffer[2] = 0xbf;
			BytesWritten = 3;
		}

		if ( WriteFile( FileHandle, UTF8Buffer, BytesWritten, &BytesWritten, NULL ) == 0 )
			rv = 2;
	}

	while ( ThisKey ) {
		GetKeyName( ThisKey->Key, KeyString );

		if ( ThisKey->Flags & FLAG_ALL_DEFINED ) {
			wcscat_s( KeyString, TextSize, L" /" );

			if ( ThisKey->Flags & FLAG_NO_ERROR_BEEP )
				wcscat_s( KeyString, TextSize, L"B" );

			if ( ThisKey->Flags & FLAG_COMMAND_PREFIX )
				wcscat_s( KeyString, TextSize, L"C" );

			if ( ThisKey->Flags & FLAG_AUTO_ENTER )
				wcscat_s( KeyString, TextSize, L"E" );

			if ( ThisKey->Flags & FLAG_INSERT )
				wcscat_s( KeyString, TextSize, L"I" );

			if ( ThisKey->Flags & FLAG_AUTOQUOTE )
				wcscat_s( KeyString, TextSize, L"Q" );

			if ( ThisKey->Flags & FLAG_CHANGE_OPTION )
				wcscat_s( KeyString, TextSize, L"T" );

			if ( ThisKey->Flags & FLAG_STRIP_QUOTES )
				wcscat_s( KeyString, TextSize, L"U" );

			if ( ThisKey->Flags & FLAG_EXPAND_VARIABLES )
				wcscat_s( KeyString, TextSize, L"X" );
		}

		wcscat_s( KeyString, TextSize, L" = " );
		wcscat_s( KeyString, TextSize, &(ThisKey->Text) );
		
		if ( ToClipboard || ToStdout )
			wcscat_s( KeyString, TextSize, L"\r\n" );
		else
			wcscat_s( KeyString, TextSize, L"\n" );

		if ( ToStdout ) {
			QPuts( KeyString );
			error = 1;
		}
		else if ( UTF16 )
			error = WriteFile( FileHandle, KeyString, (DWORD) ( wcslen( KeyString ) * sizeof( wchar_t ) ), &BytesWritten, NULL );
		else {
			WideCharToMultiByte( CP_UTF8, 0, KeyString, -1, (LPSTR) UTF8Buffer, 2 * TextSize, NULL, NULL );
			error = WriteFile( FileHandle, UTF8Buffer, (DWORD) strlen( (LPSTR) UTF8Buffer ), &BytesWritten, NULL );
		}

		if ( error == 0 ) {
			if ( Complain ) {
				GetErrorText( GetLastError(), ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error writing to file %s :\r\n", Filename );
				Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
			}

			rv = 2;
			break;
		}

		ThisKey = ThisKey->Next;
	}
	
	if ( FileHandle )
		CloseHandle( FileHandle );


	if ( ToClipboard && ( rv == 0 ) ) {
		if ( PutFileOnClipboard( Filename, TRUE ) ) {
			if ( Complain ) {
				GetErrorText( GetLastError(), ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error copying to clipboard: %s\r\n", ErrorText );
			}

			rv = 2;
		}
	}

	if ( DeleteWhenDone )
		DeleteFile( Filename );

	return rv;
}



//	Display a single key definition to stdout.
//	MinLength:  minimum length for the key name and option flags

int ShowKeyEntry( _KeyEntry *KeyEntry, int MinLength )
{
	if ( KeyEntry == NULL )
		return -1;

	wchar_t shortbuf[MAX_ARG_LEN] = L"";

	GetKeyName( KeyEntry->Key, shortbuf );

	if ( KeyEntry->Flags & FLAG_NO_ERROR_BEEP )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /B" );

	if ( KeyEntry->Flags & FLAG_COMMAND_PREFIX )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /C" );

	if ( KeyEntry->Flags & FLAG_AUTO_ENTER )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /E" );

	if ( KeyEntry->Flags & FLAG_INSERT )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /I" );

	if ( KeyEntry->Flags & FLAG_AUTOQUOTE )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /Q" );

	if ( KeyEntry->Flags & FLAG_CHANGE_OPTION )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /T" );

	if ( KeyEntry->Flags & FLAG_STRIP_QUOTES )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /U" );

	if ( KeyEntry->Flags & FLAG_EXPAND_VARIABLES )
		wcscat_s( shortbuf, MAX_ARG_LEN, L" /X" );

	if ( (int) wcslen( shortbuf ) < MinLength ) {
		for ( int i = (int) wcslen( shortbuf ); i < MinLength; i++ )
			shortbuf[i] = 0x20;

		shortbuf[MinLength] = '\0';
	}

	wcscat_s( shortbuf, MAX_ARG_LEN, L" = " );
	wcscat_s( shortbuf, MAX_ARG_LEN, &(KeyEntry->Text) );
	wcscat_s( shortbuf, MAX_ARG_LEN, L"\r\n" );

	QPuts( shortbuf );
	return 0;
}



//	Dump all key definitions to stdout.

int ShowAllKeyEntries()
{
	_KeyEntry *ThisEntry = FirstKey;

	if ( ThisEntry == NULL ) {
		QPuts( L"No keys defined.\r\n" );
		return 0;
	}

	int Length = 0;
	int Longest = 0;
	wchar_t KeyName[SHORT_BUF_LEN] = L"";
	int n = 0;


	//  Find the longest keyname-and-options string so we can pretty-print entries:

	while ( ThisEntry ) {
		GetKeyName( ThisEntry->Key, KeyName );

		Length = (int) wcslen( KeyName );

		if ( ThisEntry->Flags & FLAG_NO_ERROR_BEEP )
			Length += 3;

		if ( ThisEntry->Flags & FLAG_AUTO_ENTER )
			Length += 3;

		if ( ThisEntry->Flags & FLAG_INSERT )
			Length += 3;

		if ( ThisEntry->Flags & FLAG_STRIP_QUOTES )
			Length += 3;

		if ( ThisEntry->Flags & FLAG_AUTOQUOTE )
			Length += 3;

		if ( ThisEntry->Flags & FLAG_EXPAND_VARIABLES )
			Length += 3;

		if ( ThisEntry->Flags & FLAG_COMMAND_PREFIX )
			Length += 3;

		if ( ThisEntry->Flags & FLAG_CHANGE_OPTION )
			Length += 3;

		if ( Length > Longest )
			Longest = Length;

		ThisEntry = ThisEntry->Next;
	}

	//	Display all key definitions:

	ThisEntry = FirstKey;

	while ( ThisEntry ) {
		ShowKeyEntry( ThisEntry, Longest );
		ThisEntry = ThisEntry->Next;
		n++;
	}

	return n;
}



//	Find the definition for a key value, and delete it from the list :

int DeleteKeyEntry( int KeyValue )
{
	if ( FirstKey == NULL )
		return 1;

	_KeyEntry *ThisKey = FirstKey, *LastKey = FirstKey;

	if ( FirstKey->Key == KeyValue ) {
		ThisKey = FirstKey->Next;
		free( FirstKey );
		FirstKey = ThisKey;
		return 0;
	}

	ThisKey = ThisKey->Next;

	do {
		if ( ThisKey->Key == KeyValue ) {
			LastKey->Next = ThisKey->Next;
			free( ThisKey );
			return 0;
		}

		LastKey = ThisKey;
		ThisKey = ThisKey->Next;

	} while ( ThisKey );

	return 1;
}



//	Find the definition for a key value, and return its address :
//	Returns NULL if no definition found.

_KeyEntry* FindKeyEntry( int KeyValue )
{
	_KeyEntry *ThisKey = FirstKey;

	while ( ThisKey ) {
		if ( ThisKey->Key == KeyValue )
			return ThisKey;

		ThisKey = ThisKey->Next;
	}

	return NULL;
}



//	Finds and deletes key definitions for keynames matching a wildspec.

int DeleteMatchingKeyEntries( LPTSTR Wildspec )
{
	if ( !Wildspec )
		return -1;
	else if ( Wildspec[0] == '\0' )
		return 0;


	_KeyEntry *ThisEntry = FirstKey, *NextEntry = NULL, *PrevEntry = NULL;
	wchar_t KeyName[SHORT_BUF_LEN] = L"";
	int Count = 0;


	while ( ThisEntry ) {
		NextEntry = ThisEntry->Next;
		GetKeyName( ThisEntry->Key, KeyName );

		if ( WildcardComparison( Wildspec, KeyName, FALSE, TRUE ) == 0 ) {
			if ( PrevEntry )
				PrevEntry->Next = NextEntry;
			else
				FirstKey = NextEntry;

			free( ThisEntry );
			Count++;
		}
		else
			PrevEntry = ThisEntry;

		ThisEntry = NextEntry;
	}

	return Count;
}



/*
_KeyEntry* FindKeyEntryByIndex( unsigned int Index )
{
	_KeyEntry *ThisKey = FirstKey;

	for ( unsigned int i = 0; i < Index; i++ )
		if ( ThisKey )
			ThisKey = ThisKey->Next;

	return ThisKey;
}
*/



//	Creates a new key definition and adds it to the list; deletes any
//	old definition for the key.  If Text is null, just deletes any
//	old definition.

int SetKeyEntry( int KeyValue, unsigned int Flags, LPTSTR Text )
{
	if ( KeyValue == 0 )
		return -1;
	
	if ( !Text ) {
		DeleteKeyEntry( KeyValue );
		return 0;
	}

	size_t len = wcslen( Text );

	_KeyEntry *NewKeyEntry = NULL;
	_KeyEntry *ThisEntry = FirstKey;
	
	NewKeyEntry = (_KeyEntry*) malloc( sizeof( _KeyEntry ) + ( len * sizeof( wchar_t ) ) );

	if ( NewKeyEntry == NULL )
		return 1;

	if ( Flags & FLAG_CHANGE_OPTION )
		Flags &= ( FLAG_CHANGE_OPTION | FLAG_EXPAND_VARIABLES | FLAG_NO_ERROR_BEEP );

	NewKeyEntry->Key		= KeyValue;
	NewKeyEntry->Flags		= Flags;
	NewKeyEntry->Next		= NULL;
	NewKeyEntry->Text		= '\0';

	if ( Text )
		wcscpy_s( &(NewKeyEntry->Text), sizeof( wchar_t ) * ( len + 1 ), Text );
	
	if ( FirstKey == NULL ) {
		FirstKey = NewKeyEntry;
		return 0;
	}
	else if ( KeyValue == FirstKey->Key ) {
		NewKeyEntry->Next = FirstKey->Next;
		free( FirstKey );
		FirstKey = NewKeyEntry;
		return 0;
	}

	while ( ThisEntry->Next ) {
		if ( KeyValue != (ThisEntry->Next)->Key )
			ThisEntry = ThisEntry->Next;
		else
			break;
	}

	NewKeyEntry->Next = ThisEntry->Next;
	ThisEntry->Next = NewKeyEntry;

	if ( NewKeyEntry->Next ) {
		if ( (NewKeyEntry->Next)->Key == KeyValue ) {
			_KeyEntry *t = (NewKeyEntry->Next)->Next;
			free( NewKeyEntry->Next );
			NewKeyEntry->Next = t;
		}
	}

	return 0;
}



BOOL StupidEqualsFix( LPTSTR Line )
{
	if ( !Line )
		return FALSE;
	else if ( Line[0] == '\0' )
		return FALSE;


	wchar_t *ch			= wcsstr( Line, L"-=" );
	int rv				= FALSE;


	while ( ch ) {
		if ( ch > Line ) {
			if ( iswalpha( *(ch - 1) ) ) {
				*(ch + 1) = FULLWIDTH_EQUALS;
				rv = TRUE;
			}
		}

		ch = wcsstr( ch + 2, L"-=" );
	}

	return rv;
}



//	Loads key definitions from a file.
//	Filename:  may be CLIP: (load from clipboard) or - (read from stdin)
//	Complain:  display error messages to stderr
//	Verbose:   display the number of key definitions read to stdout when done
//	Returns 0 for success, nonzero for error.

int LoadKeysFromFile( LPTSTR Filename, BOOL Complain, BOOL Verbose )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t LineBuffer[MAX_ARG_LEN]		= L"";
	size_t BufferIndex					= 0;
	int rv								= 0;
	_MMFile MMFile;

	BOOL eqfix							= FALSE;
	int KeyValue						= 0;
	unsigned int Flags					= 0;
	int i								= 0;
	int Count							= 0;
	int ske								= 0;


	rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES, GetOEMCP(), 0 );

	if ( rv == ERROR_EMPTY_FILE )
		return 0;
	else if ( rv )
		return 2;


	while ( BufferIndex <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) )
			break;

		eqfix = StupidEqualsFix( LineBuffer );

		ParseArgs( LineBuffer, PARSE_BREAK_SPACES | PARSE_EQUALS_KLUDGE | PARSE_EQ_QT_KLUDGE );

		if ( ( arg[0][0] == ';' ) || ( arg[0][0] == ':' ) )
			continue;

		if ( eqfix )
			ReplaceChar( arg[0], FULLWIDTH_EQUALS, '=' );


		KeyValue = ParseKeyName( arg[0] );

		if ( KeyValue <= 0 ) {
			if ( Complain ) {
				_wcsupr_s( arg[0], MAX_ARG_LEN );
				DisplayErrorHeader();

				if ( KeyValue == -1 )
					Qprintf( ERRHANDLE, L"%s - Invalid key: %s\r\n", Filename, arg[0] );
				else if ( KeyValue == -2 )
					Qprintf( ERRHANDLE, L"%s - Key not allowed: %s\r\n", Filename, arg[0] );
				else
					Qprintf( ERRHANDLE, L"%s - Unknown key: %s\r\n", Filename, arg[0] );
			}

			continue;
		}

		i = 1;
		Flags = 0;

		while ( arg[i][0] == '/' ) {
			if ( ParseFlagsOption( arg[i], &Flags ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"%s - Invalid flags option: %s\r\n", Filename, arg[1] );
			}

			i++;
		}

		if ( arglen[i] ) {
			if ( arg[i][0] == '=' )
				ske = SetKeyEntry( KeyValue, Flags, arg[i] + 1 );
			else
				ske = SetKeyEntry( KeyValue, Flags, arg[i] );

			if ( ske == 0 )
				Count++;
		}
	}

	CloseMMFile( &MMFile );

	if ( Verbose ) {
		if ( Count == 0 )
			QPuts( L"No hotkeys" );
		else if ( Count == 1 )
			QPuts( L"1 hotkey" );
		else
			Printf( L"%d hotkeys", Count );

		Printf( L" loaded from %s.\r\n", Filename );
	}

	return rv;
}



//	Looks for AUTO.EK and reads key definitions from it.

int AutoLoadKeys()
{
	wchar_t Filename[MAX_FILENAME_LEN]	= L"";
	wchar_t *NamePart					= NULL;


	GetModuleFileName( (HMODULE) MyDLLHandle, Filename, MAX_FILENAME_LEN );
	NamePart = _FilenamePart( Filename, NULL );
	*NamePart = '\0';
	AppendFilename( Filename, (LPTSTR) AutoLoadFilename, MAX_FILENAME_LEN );		//	MakeDirectoryName( Filename, (LPTSTR) AutoLoadFilename );
	//	wprintf( L"   * AutoLoadKeys() trying filename %s :\n", Filename );

	if ( IsAFile( Filename ) )
		return LoadKeysFromFile( Filename, FALSE, FALSE );


	wcscpy_s( Filename, MAX_FILENAME_LEN, L"%_cmdspec" );
	ExpandVariables( Filename, 1 );
	NamePart = _FilenamePart( Filename, NULL );
	*NamePart = '\0';
	AppendFilename( Filename, (LPTSTR) AutoLoadFilename, MAX_FILENAME_LEN );		//	MakeDirectoryName( Filename, (LPTSTR) AutoLoadFilename );
	//	wprintf( L"   * AutoLoadKeys() trying filename %s :\n", Filename );

	if ( IsAFile( Filename ) )
		return LoadKeysFromFile( Filename, FALSE, FALSE );

	return 0;
}



LPTSTR FirstUnquotedSpace( LPTSTR String )
{
	if ( !String )
		return NULL;


	BOOL quote = FALSE;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( ch == QUOTE )
			quote = !quote;

		if ( ( ch == SPACE ) && !quote )
			return String + i;
	}

	return NULL;
}



//	EKEY command:

DLLExports int WINAPI ekey( LPTSTR inString )
{
	if ( inString == NULL )
		return -1;


	int NewKeyValue				= 0;
	LPTSTR NewKeyText			= NULL;
	unsigned int Flags			= 0;
	BOOL eqfix					= FALSE;

	BOOL Quiet					= FALSE;
	BOOL ListKeys				= FALSE;
	BOOL Load					= FALSE;
	BOOL Save					= FALSE;
	BOOL Overwrite				= FALSE;
	BOOL UTF16					= FALSE;
	BOOL DeleteEntry			= FALSE;
	BOOL DeleteMany				= FALSE;

	wchar_t Filename[MAX_FILENAME_LEN]	= L"";
	wchar_t NewKeyName[SHORT_BUF_LEN]	= L"";



	eqfix = StupidEqualsFix( inString );

	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_EQUALS_KLUDGE | PARSE_EQ_QT_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( EKeyHelpText, ( arg[i][2] == '?' ) );
			else if ( !wcscmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( ParseFlagsOption( arg[i], &Flags ) == 0 )
				;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				DeleteEntry = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/D", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing wildspec: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( Filename, MAX_FILENAME_LEN, arg[i] + j );
				DeleteMany = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				ListKeys = TRUE;
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				Overwrite = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing filename: \"%s\"\r\n", arg[i] );
					return 1;
				}

				Load = TRUE;

				if ( wcscmp( arg[i] + j, StdinFilename ) )
					QueryTrueName( arg[i] + j, Filename );
				else
					wcscpy_s( Filename, MAX_FILENAME_LEN, arg[i] + j );
			}
			else if ( !_wcsnicmp( arg[i], L"/S", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing filename: \"%s\"\r\n", arg[i] );
					return 1;
				}

				Save = TRUE;

				if ( wcscmp( arg[i] + j, StdinFilename ) )
					QueryTrueName( arg[i] + j, Filename );
				else
					wcscpy_s( Filename, MAX_FILENAME_LEN, arg[i] + j );
			}
			else if ( !_wcsicmp( arg[i], L"/UTF16" ) )
				UTF16 = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				Quiet = TRUE;
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( NewKeyValue == 0 ) {
				if ( eqfix )
					ReplaceChar( arg[i], FULLWIDTH_EQUALS, '=' );

				NewKeyValue = ParseKeyName( arg[i] );

				if ( NewKeyValue <= 0 ) {
					DisplayErrorHeader();

					if ( NewKeyValue == -1 )
						Qprintf( ERRHANDLE, L"Invalid key name: \"%s\"\r\n", arg[i] );
					else if ( NewKeyValue == -2 )
						Qprintf( ERRHANDLE, L"Key not allowed: \"%s\"\r\n", arg[i] );
					else
						Qprintf( ERRHANDLE, L"Unknown key name: \"%s\"\r\n", arg[i] );

					return 1;
				}

				GetKeyName( NewKeyValue, NewKeyName );
			}
			else if ( NewKeyText == NULL )
				NewKeyText = arg[i];
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	if ( ListKeys ) {
		if ( NewKeyValue || Flags || Save || Load || Overwrite || DeleteEntry || DeleteMany ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /L with other options\r\n" );
			return 1;
		}

		ListLegalKeys();
		return 0;
	}

	if ( Save ) {
		if ( NewKeyValue || Flags || Load || DeleteEntry || DeleteMany ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /S: with other options\r\n" );
			return 1;
		}

		return SaveKeysToFile( Filename, Overwrite, TRUE, UTF16 );
	}

	if ( Load ) {
		if ( NewKeyValue || Flags || Save || DeleteEntry || DeleteMany ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /R: with other options\r\n" );
			return 1;
		}

		return LoadKeysFromFile( Filename, TRUE, !Quiet );
	}

	if ( DeleteMany ) {
		if ( NewKeyValue || Flags ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /Z with other options\r\n" );
			return 1;
		}

		int Count = DeleteMatchingKeyEntries( Filename );

		if ( !Quiet )
			Printf( L"%d key%s deleted.\r\n", Count, ( Count == 1 ) ? L"" : L"s" );

		return 0;
	}

	if ( DeleteEntry && ( NewKeyValue == 0 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing key name\r\n" );
		return 1;
	}

	if ( NewKeyValue == 0 ) {
		ShowAllKeyEntries();
		return 0;
	}

	if ( DeleteEntry && ( NewKeyText == NULL ) ) {
		int rv = DeleteKeyEntry( NewKeyValue );

		if ( rv == 1 ) {
			if ( !Quiet ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Key not defined: %s\r\n", NewKeyName );
			}

			return 2;
		}
		else if ( !Quiet )
			Printf( L"%s deleted.\r\n", NewKeyName );
		
		return 0;
	}

	if ( NewKeyText == NULL ) {
		_KeyEntry *KeyEntry = FindKeyEntry( NewKeyValue );

		if ( KeyEntry == NULL ) {
			if ( !Quiet ) {
				_wcsupr_s( NewKeyName, MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Key not defined: %s\r\n", NewKeyName );
			}

			return 2;
		}

		ShowKeyEntry( KeyEntry, 0 );
		return 0;
	}

	if ( NewKeyText[0] == '=' )
		NewKeyText++;

	if ( Flags & FLAG_COMMAND_PREFIX ) {		//	kludge for if the user types a command prefix with no key definition text:
		if ( FirstUnquotedSpace( NewKeyText ) == NULL ) {
			wcscpy_s( Filename, MAX_FILENAME_LEN, NewKeyText );
			wcscat_s( Filename, MAX_FILENAME_LEN, L" #" );
			NewKeyText = Filename;
		}
	}

	int rv = SetKeyEntry( NewKeyValue, Flags, NewKeyText );

	if ( rv ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error %d defining key: %s\r\n", rv, NewKeyName );
		return 2;
	}

	return 0;
}



int ExpandVariablesQuietly( LPTSTR Buffer, int Flags = 1 )
{
	if ( !Buffer )
		return -1;


	HANDLE OldStderrHandle	= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle	= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewStderrHandle	= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
	int rv	= 0;

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewStderrHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, NewStderrHandle );
	}

	rv = ExpandVariables( Buffer, Flags );

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, OldStdoutHandle );
		CloseHandle( NewStderrHandle );
	}
	
	return rv;
}



int SetShellOption( LPTSTR OptionName, LPTSTR OptionValue, BOOL ShowStatusBar )
{
	if ( !OptionName || !OptionValue )
		return -1;


	wchar_t CmdBuffer[MAX_ARG_LEN]	= L"//";

	wcscat_s( CmdBuffer, MAX_ARG_LEN, OptionName );
	wcscat_s( CmdBuffer, MAX_ARG_LEN, L"=" );
	wcscat_s( CmdBuffer, MAX_ARG_LEN, OptionValue );


	HANDLE OldStderrHandle	= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle	= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewStderrHandle	= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
	int rv = 0;

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewStderrHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, NewStderrHandle );
	}

	rv = CallInternal( L"option", CmdBuffer );			//	rv = Option_Cmd( CmdBuffer );

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );
		SetStdHandle( STD_OUTPUT_HANDLE, OldStdoutHandle );
		CloseHandle( NewStderrHandle );
	}

	if ( ShowStatusBar ) {
		if ( rv ) {
			wchar_t MsgBuf[MAX_ARG_LEN] = L"**Error**  ";
			wcscat_s( MsgBuf, CmdBuffer + 2 );
			StatusBar( MsgBuf );
		}
		else
			StatusBar( CmdBuffer + 2 );
	}

	return rv;
}



int GetOptionValue( LPTSTR OptionName, LPTSTR OptionValue )
{
	if ( !OptionName || !OptionValue )
		return -1;


	wchar_t Buffer[MAX_ARG_LEN] = L"%@option[";


	wcscat_s( Buffer, MAX_ARG_LEN, OptionName );
	wcscat_s( Buffer, MAX_ARG_LEN, L"]" );
	OptionValue[0] = '\0';

	if ( ExpandVariablesQuietly( Buffer ) )
		return 1;

	wcscpy_s( OptionValue, MAX_ARG_LEN, Buffer );
	return 0;
}



//	Toggles a value:  returns 'yes' for 'no', 'on' for 'off', '1' for '0', etc.

int ToggleOption( LPTSTR String )
{
	if ( !String )
		return -1;


	TrimLine( String );


	if ( !_wcsicmp( String, L"on" ) )
		wcscpy_s( String, MAX_ARG_LEN, L"Off" );
	else if ( !_wcsicmp( String, L"off" ) )
		wcscpy_s( String, MAX_ARG_LEN, L"On" );
	else if ( !_wcsicmp( String, L"yes" ) )
		wcscpy_s( String, MAX_ARG_LEN, L"No" );
	else if ( !_wcsicmp( String, L"no" ) )
		wcscpy_s( String, MAX_ARG_LEN, L"Yes" );
	else if ( !_wcsicmp( String, L"true" ) )
		wcscpy_s( String, MAX_ARG_LEN, L"False" );
	else if ( !_wcsicmp( String, L"false" ) )
		wcscpy_s( String, MAX_ARG_LEN, L"True" );
	else {
		LPTSTR num = String;
		unsigned int value = 0;

		if ( num[0] == '-' )
			num;
		
		if ( ParseInt( num, &value, NULL ) == 0 )
			wsprintf( String, L"%u", !value );
	}

	return 0;
}



wchar_t *findsep( wchar_t *String )
{
	if ( !String )
		return NULL;

	wchar_t *ch = String;

	while ( *ch ) {
		if ( IsColonOrEquals( *ch ) )
			return ch;
		else if ( iswspace( *ch ) )
			return ch;

		ch++;
	}

	return NULL;
}



int ChangeShellOption( LPTSTR String, unsigned int Flags )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return -1;


	wchar_t OptionName[MAX_ARG_LEN]		= L"";
	wchar_t OptionValue[MAX_ARG_LEN]	= L"";
	wchar_t *sep1						= NULL;
	int rv								= 0;


	wcscpy_s( OptionName, MAX_ARG_LEN, String );
	TrimLine( OptionName );
	sep1 = findsep( OptionName );

	if ( sep1 ) {
		wcscpy_s( OptionValue, MAX_ARG_LEN, sep1 + 1 );
		*sep1 = '\0';
	}
	else {
		GetOptionValue( OptionName, OptionValue );
		ToggleOption( OptionValue );
	}

	if ( OptionValue[1] && ( Flags & FLAG_EXPAND_VARIABLES ) )
		ExpandVariablesQuietly( OptionValue );


	if ( ( !_wcsicmp( OptionName, L"CompleteHidden" )) && ( ShellMajor >= 10 ) ) {
		if ( sep1 == NULL ) {
			GetOptionValue( L"CompleteHiddenFiles", OptionValue );
			ToggleOption( OptionValue );
		}

		rv  = SetShellOption( L"CompleteHiddenDirs",  OptionValue, FALSE );
		rv |= SetShellOption( L"CompleteHiddenFiles", OptionValue, TRUE );
	}
	else
		rv = SetShellOption( OptionName, OptionValue, TRUE );

	if ( rv && !( Flags & FLAG_NO_ERROR_BEEP ) )
		honk();

	return rv;
}



//	Modifies the command line and cursor info passed via *ki, according to the hotkey
//	definition in *KeyEntry.  May return:
//
//		-1		bad pointer / parameter error
//		0		command line not changed
//		1		command line changed

int HandleHotkey( KEYINFO24 *ki, _KeyEntry *KeyEntry )
{
	if ( !ki || !KeyEntry )
		return -1;


	ki->nKey = 0;								//	tell tcc not to handle the keystroke
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';


	//	If /T hotkey, handle it with ChangeShellOption:

	if ( KeyEntry->Flags & FLAG_CHANGE_OPTION ) {
		ChangeShellOption( &KeyEntry->Text, KeyEntry->Flags );
		return 0;
	}


	wchar_t CursorWord[MAX_CMD_LINE_LEN]		= L"";							//	the word found at the cursor
	wchar_t Temp[MAX_CMD_LINE_LEN]				= L"";							//	the text which will replace it
	wchar_t NewLine[MAX_CMD_LINE_LEN]			= L"";							//	the modified command line
	LPTSTR CommandLine							= ki->pszLine;					//	pointer to the original command line
	LPTSTR KeyText								= &(KeyEntry->Text);			//	the text of the hotkey definition
	LPTSTR Prefix								= NULL;							//	points to the /C command prefix, if any
	int PrefixLen								= 0;							//	length of the /C command prefix, if any

	int CursorPosition							= (int) ( ki->pszCurrent - ki->pszLine );

	int WordStart = -1, WordEnd = -1;
	BOOL WordFound		= FALSE;
	wchar_t ch = '\0';
	BOOL quote = FALSE, inWord = FALSE;
	BOOL Insert = ( KeyEntry->Flags & FLAG_INSERT ) != 0;


	//	If this is a /C hotkey, split the key text into the command and the remainder:

	if ( KeyEntry->Flags & FLAG_COMMAND_PREFIX ) {
		LPTSTR FirstSpace = FirstUnquotedSpace( KeyText );

		if ( FirstSpace == KeyText )
			FirstSpace = NULL;

		if ( FirstSpace ) {
			Prefix = KeyText;						//	Prefix points to the command prefix
			KeyText = FirstSpace + 1;				//	and KeyText points to the remainder of the hotkey text
			PrefixLen = (int) ( KeyText - Prefix );
		}
	}


	//	Scan the command line and find the word at the cursor:

	for ( int i = 0; CommandLine[i]; i++ ) {
		ch = CommandLine[i];

		if ( ch == QUOTE )
			quote = !quote;

		if ( !iswspace( ch ) && ( i <= CursorPosition ) ) {
			if ( !inWord )
				WordStart = i;

			inWord = TRUE;
		}

		if ( !iswspace( ch ) && inWord )
			WordEnd = i;

		if ( iswspace( ch ) && !quote )
			inWord = FALSE;
	}

	if ( WordStart >= 0 )
		if ( WordEnd >= ( CursorPosition - 1 ) )
			WordFound = TRUE;


	if ( !WordFound && !Insert && ( PrefixLen == 0 ) )
		return 0;											//	nothing to do!  command line not changed


	if ( WordFound ) {

		//	Copy the word at the cursor to the CursorWord buffer:

		for ( int i = 0; i <= ( WordEnd - WordStart ); i++ )
			CursorWord[i] = CommandLine[WordStart + i];

		CursorWord[WordEnd - WordStart + 1] = '\0';

		//	Unquote it if /U was specified:

		if ( KeyEntry->Flags & FLAG_STRIP_QUOTES )
			Unquote( CursorWord );
	}

	if ( WordFound || Insert ) {

		//	Copy the key definition to the Temp buffer, replacing number signs
		//	with CursorWord as we go:

		int j = 0;
		Temp[j] = '\0';

		for ( int i = 0; KeyText[i]; i++ ) {
			ch = KeyText[i];

			if ( ch == PARAM_CHAR ) {
				if ( KeyText[i + 1] == PARAM_CHAR ) {
					i++;

					if ( j < ( MAX_CMD_LINE_LEN - 1 ) ) {
						Temp[j++] = ch;
						Temp[j] = '\0';
					}
				}
				else {
					wcscat_s( Temp, MAX_CMD_LINE_LEN, CursorWord );
					j = (int) wcslen( Temp );
				}
			}
			else if ( ch == ZWNJ )
				;
			else if ( j < ( MAX_CMD_LINE_LEN - 1 ) ) {
				if ( ( ch - SAFECHAR_DIFF ) == PARAM_CHAR )
					ch = PARAM_CHAR;

				Temp[j++] = ch;
				Temp[j] = '\0';
			}
		}

		//	If /X was specified, expand variables and functions:

		if ( KeyEntry->Flags & FLAG_EXPAND_VARIABLES ) {
			int rv = ExpandVariablesQuietly( Temp );

			if ( rv ) {
				if ( ( KeyEntry->Flags & FLAG_NO_ERROR_BEEP ) == 0 )
					honk();

				return 0;		//	error expanding variables - command line not changed
			}
		}

		//	If /Q was specified and the string contains funky characters, quote it:

		if ( KeyEntry->Flags & FLAG_AUTOQUOTE )
			AutoQuote( Temp, MAX_CMD_LINE_LEN );

		//	If debug mode is on, prompt the user:

		if ( Debug )
			if ( FmtMsgBox( TRUE, PLUGIN_NAME, L"Replace:\t%s  \nWith:\t%s  \n", CursorWord, Temp ) == 0 )
				return 0;		//	user canceled ekey - command line not changed
	}


	//	Empty NewLine:

	int j = 0;
	NewLine[j] = '\0';


	if ( WordFound ) {
		//	Find the next character after the cursor word; this is the
		//	remainder of the command line to copy to the new command line:

		int Following = WordEnd + 1;

		if ( ( Temp[0] == '\0' ) && CommandLine[Following] )
			while ( iswspace( CommandLine[Following] ) )
				Following++;

		//	Add everything before the cursor word to NewLine:

		for ( int i = 0; i < WordStart; i++ )
			if ( j < ( MAX_CMD_LINE_LEN - 1 ) )
				NewLine[j++] = CommandLine[i];

		//	Add the contents of the Temp buffer to NewLine:

		for ( int i = 0; Temp[i]; i++ )
			if ( j < ( MAX_CMD_LINE_LEN - 1 ) )
				NewLine[j++] = Temp[i];

		//	Add everything after the cursor word to NewLine:

		for ( int i = Following; CommandLine[i]; i++ )
			if ( j < ( MAX_CMD_LINE_LEN - 1 ) )
				NewLine[j++] = CommandLine[i];

		NewLine[j] = '\0';
	}
	else if ( Insert ) {
		//	Copy in everything before the cursor:

		for ( int i = 0; i < CursorPosition; i++ )
			if ( j < ( MAX_CMD_LINE_LEN - 1 ) )
				NewLine[j++] = CommandLine[i];

		//	Add the contents of the Temp buffer to NewLine:

		for ( int i = 0; Temp[i]; i++ )
			if ( j < ( MAX_CMD_LINE_LEN - 1 ) )
				NewLine[j++] = Temp[i];

		//	Add everything after the cursor to NewLine:

		for ( int i = CursorPosition; CommandLine[i]; i++ )
			if ( j < ( MAX_CMD_LINE_LEN - 1 ) )
				NewLine[j++] = CommandLine[i];

		NewLine[j] = '\0';
	}
	else
		wcscpy_s( NewLine, MAX_CMD_LINE_LEN, CommandLine );


	//	If there is a /C command, prefix it to the start of the command line:

	if ( Prefix ) {
		wchar_t *Start = NewLine;				//	scan the modified command line
		if ( *Start )
			while ( iswspace( *Start ) )		//	for the first nonspace character
				Start++;

		if ( _wcsnicmp( Start, Prefix, PrefixLen ) ) {							//	if our command prefix isn't already there,
			if ( ( wcslen( NewLine ) + PrefixLen ) < MAX_CMD_LINE_LEN ) {		//	and if there is room,
				for ( size_t i = wcslen( NewLine ); i != ( (size_t) -1 ); i-- )
					NewLine[PrefixLen + i] = NewLine[i];						//	move the new command line up to make space

				for ( int i = 0; i < PrefixLen; i++ ) {							//	then copy in the command prefix
					ch = Prefix[i];
					if ( ch == NBSP )											//	replacing non-breaking spaces
						ch = SPACE;												//	with regular ascii spaces

					NewLine[i] = ch;
				}
			}
			else																//	if there's not enough room,
				PrefixLen = 0;													//	it didn't happen
		}
		else																	//	if new command line already starts with the prefix,
			PrefixLen = 0;														//	it didn't happen
	}


	ClearEditLine( CommandLine );								//	so we don't leave bits of the old command line on the screen
	wcscpy_s( CommandLine, MAX_CMD_LINE_LEN, NewLine );			//	now copy in the modified command line

	if ( WordFound )											//	and update the cursor position appropriately
		ki->pszCurrent = ki->pszLine + WordStart + (int) wcslen( Temp ) + PrefixLen;
	else if ( Insert )
		ki->pszCurrent = ki->pszCurrent + (int) wcslen( Temp ) + PrefixLen;
	else if ( PrefixLen )
		ki->pszCurrent += PrefixLen;
	
	ki->fRedraw = 1;											//	tcc must redraw the command line

	if ( KeyEntry->Flags & FLAG_AUTO_ENTER ) {					//	if /E was specified,
		if ( NewKeyHook )
			wcscpy_s( ki->pszKey, SHORT_BUF_LEN, L"Enter" );	//	pretend the user pressed Enter
		else
			ki->nKey = ENTER;
	}

	return 1;													//	command line was changed
}



//	Plugin hotkey hook:  Check for defined key combination, pass to handler if found.

DLLExports int WINAPI key( LPKEYINFO24 ki )
{
	if ( !FirstKey )
		return 0;

	if ( ki->nKey == 0 ) {
		if ( !NewKeyHook )
			return 0;
		else if ( ki->pszKey == NULL )
			return 0;
	}


	int inKey = ki->nKey;
	int ShiftKeys = 0;
	_KeyEntry *KeyEntry = FirstKey;


	if ( NewKeyHook ) {
		if ( !inKey && ki->pszKey ) {
			inKey = ParseKeyName24( ki->pszKey );

			if ( !inKey )
				return 0;
		}
	}


	if ( GetAsyncKeyState( VK_SHIFT ) & 0x8000 )
		ShiftKeys |= 1;

	if ( GetAsyncKeyState( VK_CONTROL ) & 0x8000 )
		ShiftKeys |= 2;

	if ( GetAsyncKeyState( VK_MENU ) & 0x8000 )
		ShiftKeys |= 4;

	if ( ShiftKeys == 3 )
		inKey |= 0x20000;
	else if ( ShiftKeys == 6 )
		inKey |= 0x40000;
	else if ( NewKeyHook && ( ShiftKeys == 7 ) )
		inKey |= 0x6000;


	while ( KeyEntry ) {
		if ( inKey == KeyEntry->Key )
			HandleHotkey( ki, KeyEntry );

		KeyEntry = KeyEntry->Next;
	}
	
	return 0;
}



/*
int SafeChars( LPTSTR String )
{
	if ( String == NULL )
		return -1;

	int k = 0;

	for ( int i = 0; String[i]; i++ ) {
		if ( ( String[i] == QUOTE ) || ( String[i] == '%' ) || ( String[i] == '[' ) || ( String[i] == ']' ) || ( String[i] == '(' ) || ( String[i] == ')' ) || ( String[i] == '^' ) || ( String[i] == '`' ) || ( String[i] == ',' ) ) {
			String[i] += SAFECHAR_DIFF;
			k++;
		}
	}

	return k;
}


DLLExports int WINAPI f_ekey( LPTSTR String )
{
	if ( String == NULL )
		return -1;

	int KeyValue			= 0;
	unsigned int KeyIndex	= 0;
	unsigned int Which		= 0;
	BOOL Numeric			= FALSE;
	_KeyEntry *KeyEntry		= NULL;


	ParseArgs( String, PARSE_BREAK_COMMAS );

	if ( arglen[0] ) {
		if ( iswdigit( arg[0][0] ) ) {
			if ( ParseInt( arg[0], &KeyIndex, NULL ) ) {
				_wcsupr_s( arg[0], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad index: \"%s\"\r\n", arg[0] );
				return -1;
			}

			Numeric = TRUE;
		}
		else {
			KeyValue = ParseKeyName( arg[0] );

			if ( KeyValue == 0 ) {
				_wcsupr_s( arg[0], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad key name: \"%s\"\r\n", arg[0] );
				return -1;
			}
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing key name\r\n", arg[0] );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Which, NULL ) ) {
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad integer: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( Numeric )
		KeyEntry = FindKeyEntryByIndex( KeyIndex );
	else
		KeyEntry = FindKeyEntry( KeyValue );

	String[0] = '\0';

	switch ( Which & 0xff ) {

	case 1 :							//	key code

		if ( KeyEntry )
			GetKeyName( KeyEntry->Key, String );
		
		break;

	case 2 :							//  flags
		
		if ( KeyEntry )
			wsprintf( String, L"%u", KeyEntry->Flags );

		break;

	default :							//  key text

		if ( KeyEntry ) {
			wcscpy_s( String, MAX_ARG_LEN, &(KeyEntry->Text) );

			if ( Which & 0x0100 )
				SafeChars( String );
		}
	}
	
	return 0;
}



DLLExports int WINAPI _ekeys( LPTSTR outString )
{
	if ( outString == NULL )
		return -1;

	_KeyEntry *ThisKey = FirstKey;
	int n = 0;

	while ( ThisKey ) {
		ThisKey = ThisKey->Next;
		n++;
	}

	wsprintf( outString, L"%d", n );
	return 0;
}
*/



//	Plugin help hook:  Check for plugin keywords, display a message box if found.

DLLExports int WINAPI Help( LPTSTR inString )
{
	if ( !_wcsicmp( inString, L"ekey" ) ) {
		MessageBox( GetForegroundWindow(), EKeyHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	/*
	else if ( !_wcsicmp( inString, L"@ekey" ) ) {
		MessageBox( GetForegroundWindow(), F_EKeyHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_ekeys" ) ) {
		MessageBox( GetForegroundWindow(), _EKeysHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	*/
	else
		return 0;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



//	Plugin hook for usage text:  Check for plugin commands, provide text for TC's status bar if found.

DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"ekey" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( EKeyHelpText ) );
		return 1;
	}
	else
		return 0;
}



/*

	0.50.0.1	2021-09-14

	0.50.1.0	2021-12-21
				Fixed an issue with calling TakeCommandIPC() from 64-bit TCC.
				Help text is now displayed via ShowCmdHelp().

	0.50.2.0	2021-12-22
				No longer link to TakeCommandIPC() dynamically.  That hasn't been
				necessary for a long long time.

	0.50.3.0	2022-12-21

	0.50.4.0	2023-06-16
				Removed all calls to QueryIsFile() to work around an issue
				in TCC v30.00.20.

	0.51.0.0.	2023-07-25

	0.51.0.1	2023-10-12
				Updated the plugin's web address.

	0.51.0.3	2023-12-01
				Updated to the current version of mmfiles.cpp.  Cosmetic:
				Replaced a bunch of ASCII apostrophes with fancy Unicode quotes.

*/
